from __future__ import annotations

from collections import deque
import socket
import ssl
from threading import Condition
from typing import Any, cast
from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection import Address
from ohmqtt.connection.decoder import IncrementalDecoder, ClosedSocketError
from ohmqtt.connection.fsm import FSM
from ohmqtt.connection.keepalive import KeepAlive
from ohmqtt.connection.selector import InterruptibleSelector
from ohmqtt.connection.states import (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ConnectedState,
    ReconnectWaitState,
    ClosingState,
    ClosedState,
    ShutdownState,
    WebsocketHandshakeRequestState,
    WebsocketHandshakeResponseState,
)
from ohmqtt.connection.timeout import Timeout
from ohmqtt.connection.types import ConnectParams, StateData, StateEnvironment
from ohmqtt.connection.ws_decoder import WebsocketDecoder
from ohmqtt.connection.wslib import (
    OpCode,
    apply_mask,
    frame_ws_data,
    generate_nonce,
    generate_handshake_key,
    WebsocketError,
)
from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode
from ohmqtt.packet import (
    decode_packet,
    MQTTPacket,
    MQTTConnectPacket,
    MQTTConnAckPacket,
    MQTTDisconnectPacket,
    MQTTPublishPacket,
    MQTTPingReqPacket,
    MQTTPingRespPacket,
    MQTTSubscribePacket,
    MQTTUnsubscribePacket,
    PING,
)
from ohmqtt.platform import HAS_AF_UNIX
from ohmqtt.property import MQTTConnectProps, MQTTConnAckProps, MQTTWillProps


class EnvironmentCallbacks:
    """Container for StateEnvironment callbacks."""
    def __init__(self, mocker: MockerFixture) -> None:
        self.packet = mocker.Mock()

    def reset(self) -> None:
        self.packet.reset_mock()

    def assert_not_called(self) -> None:
        self.packet.assert_not_called()


@pytest.fixture
def callbacks(mocker: MockerFixture) -> EnvironmentCallbacks:
    return EnvironmentCallbacks(mocker)

@pytest.fixture
def env(callbacks: EnvironmentCallbacks) -> StateEnvironment:
    return StateEnvironment(
        packet_callback=callbacks.packet,
    )

@pytest.fixture
def mock_socket(mocker: MockerFixture) -> Mock:
    mock_socket = mocker.Mock(spec=ssl.SSLSocket)
    mocker.patch("ohmqtt.connection.states.connecting._get_socket", return_value=mock_socket)
    return mock_socket  # type: ignore[no-any-return]

@pytest.fixture
def decoder(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=IncrementalDecoder)  # type: ignore[no-any-return]

@pytest.fixture
def ws_decoder(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=WebsocketDecoder)  # type: ignore[no-any-return]

@pytest.fixture
def mock_keepalive(mocker: MockerFixture) -> Mock:
    mock_keepalive = mocker.Mock(spec=KeepAlive)
    mock_keepalive.get_next_timeout.return_value = 1
    mock_keepalive.should_close.return_value = False
    mock_keepalive.should_send_ping.return_value = False
    return mock_keepalive  # type: ignore[no-any-return]

@pytest.fixture
def mock_timeout(mocker: MockerFixture) -> Mock:
    mock_timeout = mocker.Mock(spec=Timeout)
    mock_timeout.interval = None
    mock_timeout.get_timeout.return_value = 1
    mock_timeout.exceeded.return_value = False
    return mock_timeout  # type: ignore[no-any-return]

@pytest.fixture
def state_data(decoder: Mock, mock_socket: Mock, mock_keepalive: Mock, mock_timeout: Mock, ws_decoder: Mock) -> StateData:
    data = StateData()
    data.decoder = decoder
    data.keepalive = mock_keepalive
    data.sock = mock_socket
    data.timeout = mock_timeout
    data.ws_decoder = ws_decoder
    data.connack = MQTTConnAckPacket()
    return data

@pytest.fixture
def params() -> ConnectParams:
    return ConnectParams(address=Address("mqtt://testhost"))

@pytest.fixture
def mock_select(mocker: MockerFixture) -> Mock:
    mock_select = mocker.MagicMock(spec=InterruptibleSelector)
    mock_select.__enter__.return_value = mock_select
    mock_select.select.return_value = (False, False)
    mocker.patch("ohmqtt.connection.fsm.InterruptibleSelector", return_value=mock_select)
    return mock_select  # type: ignore[no-any-return]

@pytest.fixture
def mock_read(mocker: MockerFixture) -> Mock:
    mock_read = mocker.patch.object(ConnectedState, "read_packet", autospec=True)
    mock_read.return_value = False
    return mock_read


@pytest.mark.parametrize("address", [
    "mqtt://testhost", "mqtts://testhost",
    "ws://testhost", "wss://testhost",
    "unix:///testpath",
])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connecting_happy_path(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock
) -> None:
    if address.startswith("unix:") and not HAS_AF_UNIX:
        pytest.skip("Unix socket not supported on this platform")
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=ConnectingState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    ConnectingState.enter(fsm, state_data, env, params)
    if params.address.scheme != "unix":
        mock_socket.setsockopt.assert_called_once_with(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    mock_socket.reset_mock()
    decoder.reset.assert_called_once()
    decoder.reset_mock()
    assert mock_timeout.interval == params.connect_timeout
    mock_timeout.mark.assert_called_once()
    mock_timeout.reset_mock()
    assert state_data.disconnect_rc is None
    mock_select.change_sock.assert_called_once_with(mock_socket)
    assert fsm.state is ConnectingState

    # First handle, begin non-blocking connect.
    mock_select.select.return_value = (False, False)
    mock_socket.connect.side_effect = BlockingIOError
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    if params.address.scheme == "unix":
        mock_socket.connect.assert_called_once_with(params.address.host)
    else:
        mock_socket.connect.assert_called_once_with((params.address.host, params.address.port))
    mock_socket.reset_mock()
    assert fsm.state is ConnectingState

    # Second handle, finish non-blocking connect.
    mock_select.select.return_value = (False, True)
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.setblocking.assert_called_once_with(False)
    mock_socket.reset_mock()
    if params.address.use_tls:
        assert fsm.state is TLSHandshakeState
    elif params.address.is_websocket():
        assert fsm.state is WebsocketHandshakeRequestState
    else:
        assert fsm.state is MQTTHandshakeConnectState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "mqtts://testhost", "unix:///testpath"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connecting_timeout(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_timeout: Mock
) -> None:
    if address.startswith("unix:") and not HAS_AF_UNIX:
        pytest.skip("Unix socket not supported on this platform")
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=ConnectingState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    mock_timeout.exceeded.return_value = True
    ConnectingState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_called_once()
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "mqtts://testhost", "unix:///testpath"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connecting_error(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    if address.startswith("unix:") and not HAS_AF_UNIX:
        pytest.skip("Unix socket not supported on this platform")
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=ConnectingState, error_state=ShutdownState)

    mock_socket.connect.side_effect = ConnectionError("TEST")
    ConnectingState.enter(fsm, state_data, env, params)
    ret = ConnectingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtts://testhost", "wss://testhost"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_tls_handshake_happy_path(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    mocker: MockerFixture
) -> None:
    params = ConnectParams(address=Address(address), tls_context=mocker.Mock())
    fsm = FSM(env=env, init_state=TLSHandshakeState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    params.tls_context.wrap_socket.return_value = mock_socket  # type: ignore[union-attr]
    TLSHandshakeState.enter(fsm, state_data, env, params)
    params.tls_context.wrap_socket.assert_called_once_with(  # type: ignore[union-attr]
        mock_socket,
        server_hostname=params.address.host,
        do_handshake_on_connect=False,
    )
    mock_select.change_sock.assert_called_once_with(mock_socket)

    # First handle, want write.
    mock_socket.do_handshake.side_effect = ssl.SSLWantWriteError
    mock_select.select.return_value = (False, True)
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.do_handshake.assert_called_once()
    mock_socket.reset_mock()
    assert fsm.state is TLSHandshakeState

    # Second handle, want read.
    mock_socket.do_handshake.side_effect = ssl.SSLWantReadError
    mock_select.select.return_value = (True, False)
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.do_handshake.assert_called_once()
    mock_socket.reset_mock()
    assert fsm.state is TLSHandshakeState

    # Third handle, complete.
    mock_socket.do_handshake.side_effect = None
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_socket.do_handshake.assert_called_once()
    mock_socket.reset_mock()
    if params.address.is_websocket():
        assert fsm.state is WebsocketHandshakeRequestState
    else:
        assert fsm.state is MQTTHandshakeConnectState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_tls_handshake_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mocker: MockerFixture,
    mock_select: Mock,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtts://testhost"), tls_context=mocker.Mock())
    fsm = FSM(env=env, init_state=TLSHandshakeState, error_state=ShutdownState)

    mock_timeout.exceeded.return_value = True
    TLSHandshakeState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_not_called()
    ret = TLSHandshakeState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


def test_states_tls_handshake_default_context_options(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mocker: MockerFixture
) -> None:
    params = ConnectParams(address=Address("mqtts://testhost"), tls_context=None)
    fsm = FSM(env=env, init_state=TLSHandshakeState, error_state=ShutdownState)
    state_data.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    TLSHandshakeState.enter(fsm, state_data, env, params)
    tls_context = cast(ssl.SSLSocket, state_data.sock).context
    assert tls_context.minimum_version == ssl.TLSVersion.TLSv1_3

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_happy_path(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    mocker: MockerFixture
) -> None:
    address = "mqtt://testhost"
    mock_write_buffer = mocker.Mock()
    state_data.write_buffer = mock_write_buffer
    params = ConnectParams(
        address=Address(address),
        client_id="test_client",
        keepalive_interval=60,
        clean_start=True,
        username="johndoe",
        password=b"secret",
        will_topic="test_topic",
        will_payload=b"test_payload",
        will_qos=1,
        will_retain=True,
        will_properties=MQTTWillProps(),
        connect_properties=MQTTConnectProps(),
    )
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    mock_write_buffer.clear.assert_called_once()
    mock_write_buffer.extend.assert_called_once()
    data = mock_write_buffer.extend.call_args[0][0]
    mock_write_buffer.reset_mock()
    state_data.write_buffer = bytearray(data)
    if params.address.is_websocket():
        opcode = data[0] & 0x0F
        assert opcode == OpCode.BINARY
        length = data[1] & 0x7F
        payload_offset = 2
        if length == 126:
            payload_offset = 4
        elif length == 127:
            payload_offset = 10
        mask = data[payload_offset:payload_offset + 4]
        masked_payload = data[payload_offset + 4:]
        unmasked_payload = bytearray(b ^ mask[i % 4] for i, b in enumerate(masked_payload))
        packet = decode_packet(bytes(unmasked_payload))
    else:
        packet = decode_packet(data)
    assert isinstance(packet, MQTTConnectPacket)
    assert packet.client_id == params.client_id
    assert packet.keep_alive == params.keepalive_interval
    assert packet.clean_start is params.clean_start
    assert packet.will_topic == params.will_topic
    assert packet.will_payload == params.will_payload
    assert packet.will_qos == params.will_qos
    assert packet.will_retain is params.will_retain
    assert packet.will_props == params.will_properties
    assert packet.properties == params.connect_properties
    assert packet.username == params.username
    assert packet.password == params.password
    assert fsm.state is MQTTHandshakeConnectState

    # First handle, blocked write.
    if params.address.use_tls:
        mock_socket.send.side_effect = ssl.SSLWantWriteError
    else:
        mock_socket.send.side_effect = BlockingIOError
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert fsm.state is MQTTHandshakeConnectState
    if max_wait:
        mock_select.select.assert_called_once_with(write=True, timeout=1)

    # Second handle, complete.
    mock_socket.send.side_effect = None
    mock_socket.send.return_value = len(data)
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert fsm.state is MQTTHandshakeConnAckState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)

    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_not_called()
    mock_timeout.exceeded.return_value = True
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_partial(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    mocker: MockerFixture
) -> None:
    mock_write_buffer = mocker.Mock()
    state_data.write_buffer = mock_write_buffer
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # One byte on the first write.
    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    data = mock_write_buffer.extend.call_args[0][0]
    mock_write_buffer.reset_mock()
    state_data.write_buffer = bytearray(data)
    mock_socket.send.side_effect = None
    mock_socket.send.return_value = 1
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert state_data.write_buffer == bytearray(data[1:])
    assert fsm.state is MQTTHandshakeConnectState

    # The rest on the second write.
    mock_socket.send.return_value = len(data) - 1
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert state_data.write_buffer == bytearray()
    assert fsm.state is MQTTHandshakeConnAckState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connect_error(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnectState, error_state=ShutdownState)

    MQTTHandshakeConnectState.enter(fsm, state_data, env, params)
    mock_socket.send.return_value = 0
    ret = MQTTHandshakeConnectState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "ws://testhost"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_happy_path(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    ws_decoder: Mock,
) -> None:
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    assert fsm.state is MQTTHandshakeConnAckState

    # First handle, blocked read.
    decoder.decode.return_value = None
    ws_decoder.decode.return_value = None
    mock_select.select.return_value = (True, False)
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    assert fsm.state is MQTTHandshakeConnAckState

    # Second handle, complete.
    connack = MQTTConnAckPacket(properties=MQTTConnAckProps(ServerKeepAlive=23))
    decoder.decode.return_value = connack
    ws_decoder.decode.return_value = (OpCode.BINARY, connack.encode())
    mock_select.select.return_value = (False, False)
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert state_data.connack == connack
    assert state_data.keepalive.keepalive_interval == connack.properties.ServerKeepAlive
    assert fsm.state is ConnectedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)

    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    mock_timeout.mark.assert_not_called()
    mock_timeout.exceeded.return_value = True
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_closed_socket(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    decoder: Mock,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)

    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    decoder.decode.side_effect = ClosedSocketError("TEST")
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_unexpected(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    decoder: Mock,
    state_data: StateData,
    env: StateEnvironment
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)

    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    decoder.decode.return_value = MQTTPublishPacket(topic="test/topic", payload=b"test_payload")
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_mqtt_connack_ws_protocol_error(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    ws_decoder: Mock,
    state_data: StateData,
    env: StateEnvironment
) -> None:
    params = ConnectParams(address=Address("ws://testhost"))
    fsm = FSM(env=env, init_state=MQTTHandshakeConnAckState, error_state=ShutdownState)

    MQTTHandshakeConnAckState.enter(fsm, state_data, env, params)
    ws_decoder.decode.return_value = (OpCode.TEXT, b"invalid data")
    ret = MQTTHandshakeConnAckState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "ws://testhost"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_happy_path(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    mock_keepalive: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_read: Mock,
    mocker: MockerFixture
) -> None:
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)
    mock_keepalive.get_next_timeout.return_value = 1 if max_wait is None else max_wait
    fsm.env.packet_buffer = deque()

    # Enter state.
    ConnectedState.enter(fsm, state_data, env, params)
    mock_keepalive.mark_init.assert_called_once()
    callbacks.packet.assert_called_once_with(state_data.connack)
    callbacks.reset()
    assert state_data.open_called is True
    assert fsm.state is ConnectedState

    # First handle, nothing to read.
    mock_select.select.return_value = (False, False)
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, write=False, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    assert fsm.state is ConnectedState

    # Second handle, send data in write buffer.
    packet = MQTTPublishPacket(topic="test/topic", payload=b"test_payload")
    fsm.env.packet_buffer.append(packet)
    mock_select.select.return_value = (False, True)
    if params.address.is_websocket():
        mock_socket.send.return_value = len(frame_ws_data(OpCode.BINARY, packet.encode()))
    else:
        mock_socket.send.return_value = len(packet.encode())
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert len(state_data.write_buffer) == 0
    mock_keepalive.mark_send.assert_called_once()
    assert fsm.state is ConnectedState

    # Third handle, read data.
    mock_select.select.return_value = (True, False)
    mock_read.side_effect = [True, False]
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(read=True, write=False, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    assert mock_read.call_count == 1

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_keepalive(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_keepalive: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    # First handle, send a PINGREQ.
    mock_keepalive.should_send_ping.return_value = True
    mock_select.select.return_value = (False, True)
    mock_socket.send.return_value = len(PING)
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_keepalive.should_send_ping.assert_called_once()
    mock_keepalive.mark_ping.assert_called_once()
    mock_keepalive.reset_mock()
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert len(state_data.write_buffer) == 0

    # Second handle, keepalive timeout.
    mock_keepalive.should_send_ping.return_value = False
    mock_keepalive.should_close.return_value = True
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("block_exc", [ssl.SSLWantReadError, ssl.SSLWantWriteError, BlockingIOError])
@pytest.mark.parametrize("fatal_exc", [BrokenPipeError, ConnectionResetError])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_send_errors(
    max_wait: float | None,
    block_exc: type[Exception],
    fatal_exc: type[Exception],
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_read: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    fsm.env.packet_buffer.append(MQTTPublishPacket(topic="test/topic", payload=b"test_payload"))
    mock_read.return_value = False
    mock_select.select.return_value = (True, True)

    mock_socket.send.side_effect = block_exc("TEST")
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_read.assert_called_once()
    mock_read.reset_mock()
    assert fsm.state is ConnectedState

    mock_select.select.return_value = (True, True)
    mock_socket.send.side_effect = fatal_exc("TEST")
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_read.assert_not_called()
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_read_closed(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_read: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    mock_select.select.return_value = (True, False)
    mock_read.side_effect = ClosedSocketError("TEST")
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_read_mqtt_error(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_read: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    mock_select.select.return_value = (True, False)
    mock_read.side_effect = MQTTError("TEST", reason_code=MQTTReasonCode.ProtocolError)
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert state_data.disconnect_rc == MQTTReasonCode.ProtocolError
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_connected_read_websocket_error(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_read: Mock,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    ConnectedState.enter(fsm, state_data, env, params)
    callbacks.reset()

    mock_select.select.return_value = (True, False)
    mock_read.side_effect = WebsocketError("TEST")
    ret = ConnectedState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert state_data.disconnect_rc is None
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


def test_states_connected_read_packet(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_keepalive: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    # Handle incomplete packet.
    decoder.decode.return_value = None
    ConnectedState.read_packet(fsm, state_data, env, params)
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    assert fsm.state is ConnectedState

    # Handle complete PUBLISH packet.
    pub_packet = MQTTPublishPacket(topic="test/topic", payload=b"test_payload")
    decoder.decode.return_value = pub_packet
    ConnectedState.read_packet(fsm, state_data, env, params)
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    callbacks.packet.assert_called_once_with(pub_packet)
    callbacks.reset()
    assert fsm.state is ConnectedState

    # Handle complete PINGRESP packet.
    pong_packet = MQTTPingRespPacket()
    decoder.decode.return_value = pong_packet
    ConnectedState.read_packet(fsm, state_data, env, params)
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    mock_keepalive.mark_pong.assert_called_once()
    callbacks.assert_not_called()
    assert fsm.state is ConnectedState

    # Handle complete PINGREQ packet.
    ping_packet = MQTTPingReqPacket()
    decoder.decode.return_value = ping_packet
    ConnectedState.read_packet(fsm, state_data, env, params)
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    assert fsm.env.packet_buffer.popleft() == MQTTPingRespPacket()
    fsm.env.packet_buffer.clear()
    callbacks.assert_not_called()
    assert fsm.state is ConnectedState

    # Handle complete DISCONNECT packet.
    dc_packet = MQTTDisconnectPacket()
    decoder.decode.return_value = dc_packet
    ConnectedState.read_packet(fsm, state_data, env, params)
    decoder.decode.assert_called_once()
    decoder.reset_mock()
    callbacks.assert_not_called()
    assert fsm.state is ClosingState

    # Handle invalid packets.
    for pt in (MQTTConnectPacket, MQTTSubscribePacket, MQTTUnsubscribePacket):
        packet: MQTTPacket = pt()
        decoder.decode.return_value = packet
        with pytest.raises(MQTTError) as excinfo:
            ConnectedState.read_packet(fsm, state_data, env, params)
        assert excinfo.value.reason_code == MQTTReasonCode.ProtocolError


def test_states_connected_read_packet_ws(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    ws_decoder: Mock,
    mock_keepalive: Mock
) -> None:
    params = ConnectParams(address=Address("ws://testhost"))
    fsm = FSM(env=env, init_state=ConnectedState, error_state=ShutdownState)

    # Handle incomplete packet.
    ws_decoder.decode.return_value = None
    ConnectedState.read_packet(fsm, state_data, env, params)
    ws_decoder.decode.assert_called_once()
    ws_decoder.reset_mock()
    assert fsm.state is ConnectedState

    # Handle complete PUBLISH packet.
    pub_packet = MQTTPublishPacket(topic="test/topic", payload=b"test_payload")
    ws_decoder.decode.return_value = OpCode.BINARY, pub_packet.encode()
    ConnectedState.read_packet(fsm, state_data, env, params)
    ws_decoder.decode.assert_called_once()
    ws_decoder.reset_mock()
    callbacks.packet.assert_called_once_with(pub_packet)
    callbacks.reset()
    assert fsm.state is ConnectedState

    # Handle complete WebSocket PING frame.
    ws_decoder.decode.return_value = OpCode.PING, b"\x88"
    ConnectedState.read_packet(fsm, state_data, env, params)
    ws_decoder.decode.assert_called_once()
    ws_decoder.reset_mock()
    assert state_data.write_buffer[0] == 0x80 | OpCode.PONG.value
    assert state_data.write_buffer[1] == 0x81
    mask = state_data.write_buffer[2:6]
    payload = state_data.write_buffer[6:]
    unmasked_payload = bytearray(b ^ mask[i % 4] for i, b in enumerate(payload))
    assert unmasked_payload == b"\x88"
    state_data.write_buffer.clear()
    callbacks.assert_not_called()
    assert fsm.state is ConnectedState

    # Handle WebSocket CLOSE frame.
    ws_decoder.decode.return_value = OpCode.CLOSE, b""
    with pytest.raises(ClosedSocketError):
        ConnectedState.read_packet(fsm, state_data, env, params)

    # Handle bad opcode in WebSocket frame.
    ws_decoder.decode.return_value = OpCode.TEXT, b"asdf"
    with pytest.raises(WebsocketError):
        ConnectedState.read_packet(fsm, state_data, env, params)


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_reconnect_wait_happy_path(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mocker: MockerFixture,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), reconnect_delay=5)
    fsm = FSM(env=env, init_state=ReconnectWaitState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait
    mock_cond = mocker.MagicMock(spec=Condition)
    mock_cond.wait.return_value = True
    mock_cond.__enter__.return_value = mock_cond
    fsm.cond = mock_cond

    ReconnectWaitState.enter(fsm, state_data, env, params)
    assert mock_timeout.interval == params.reconnect_delay
    mock_timeout.mark.assert_called_once()
    mock_timeout.reset_mock()

    # First handle, waiting.
    ret = ReconnectWaitState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_timeout.exceeded.assert_called_once()
    mock_timeout.reset_mock()
    if max_wait is None:
        mock_cond.wait.assert_called_once_with(mock_timeout.get_timeout.return_value)
        mock_cond.wait.reset_mock()
    else:
        mock_cond.wait.assert_not_called()
    assert fsm.state is ReconnectWaitState

    # Second handle, transition.
    mock_timeout.exceeded.return_value = True
    ret = ReconnectWaitState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_timeout.exceeded.assert_called_once()
    mock_cond.wait.assert_not_called()
    assert fsm.state is ConnectingState

    callbacks.assert_not_called()


def test_states_reconnect_wait_race(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mocker: MockerFixture
) -> None:
    """Check for a race condition between state changes and waiting for state changes."""
    params = ConnectParams(address=Address("mqtt://testhost"), reconnect_delay=5)
    fsm = FSM(env=env, init_state=ReconnectWaitState, error_state=ShutdownState)
    mock_cond = mocker.MagicMock(spec=Condition)
    mock_cond.wait.return_value = True
    mock_cond.__enter__.return_value = mock_cond
    fsm.cond = mock_cond

    ReconnectWaitState.enter(fsm, state_data, env, params)
    fsm.change_state(ConnectingState)
    ret = ReconnectWaitState.handle(fsm, state_data, env, params, True)
    assert ret is True

    callbacks.assert_not_called()


@pytest.mark.parametrize("init_dc_rc", [MQTTReasonCode.UnspecifiedError, None])
@pytest.mark.parametrize("address", ["mqtt://testhost", "mqtts://testhost"])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_closing_happy_path(
    init_dc_rc: MQTTReasonCode | None,
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address(address), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait
    state_data.disconnect_rc = init_dc_rc

    # Start from ConnectedState to test transition through ClosingState to ClosedState.
    fsm.previous_state = ConnectedState
    ClosingState.enter(fsm, state_data, env, params)
    assert mock_timeout.interval == params.connect_timeout
    mock_timeout.mark.assert_called_once()
    mock_timeout.reset_mock()
    assert state_data.disconnect_rc == MQTTReasonCode.NormalDisconnection if init_dc_rc is None else init_dc_rc
    if not params.address.use_tls:
        mock_socket.shutdown.assert_called_once_with(socket.SHUT_RD)
    else:
        mock_socket.shutdown.assert_not_called()
    mock_socket.reset_mock()
    assert fsm.state is ClosingState

    state_data.write_buffer.extend(b"test_data")

    # First handle, write blocked.
    mock_select.select.return_value = (False, False)
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_timeout.exceeded.assert_called_once()
    mock_timeout.reset_mock()
    assert fsm.state is ClosingState

    # Second handle, write all.
    mock_select.select.return_value = (False, True)
    mock_socket.send.return_value = len(state_data.write_buffer)
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    mock_select.select.assert_called_once_with(write=True, timeout=1 if max_wait is None else max_wait)
    mock_select.reset_mock()
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    mock_timeout.exceeded.assert_called_once()
    mock_timeout.reset_mock()
    assert fsm.state is ClosingState

    # Third handle, done writing, transition to ClosedState.
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    mock_timeout.exceeded.assert_called_once()
    mock_timeout.reset_mock()
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


# Try all states which can transition to ClosingState, except ConnectedState.
@pytest.mark.parametrize("prev_state", [
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    ReconnectWaitState,
])
def test_states_closing_skip(
    prev_state: Any,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = prev_state
    ClosingState.enter(fsm, state_data, env, params)
    mock_socket.shutdown.assert_not_called()
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_closing_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = ConnectedState
    ClosingState.enter(fsm, state_data, env, params)
    mock_timeout.exceeded.return_value = True
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


def test_states_closing_shutdown_error(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = ConnectedState
    mock_socket.shutdown.side_effect = OSError("TEST")
    ClosingState.enter(fsm, state_data, env, params)
    mock_socket.shutdown.assert_called_once()
    assert fsm.state is ClosingState

    callbacks.assert_not_called()


@pytest.mark.parametrize("exc", [ssl.SSLWantReadError, ssl.SSLWantWriteError, BlockingIOError])
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_closing_send_error(
    max_wait: float | None,
    exc: type[Exception],
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=ClosingState, error_state=ShutdownState)

    fsm.previous_state = ConnectedState
    ClosingState.enter(fsm, state_data, env, params)

    state_data.write_buffer.extend(b"test_data")
    mock_select.select.return_value = (False, True)
    mock_socket.send.side_effect = exc("TEST")
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    assert fsm.state is ClosingState

    mock_socket.send.side_effect = BrokenPipeError("TEST")
    ret = ClosingState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("address", ["mqtt://testhost", "ws://testhost"])
@pytest.mark.parametrize("open_called", [True, False])
@pytest.mark.parametrize("rc", [None, MQTTReasonCode.NormalDisconnection])
@pytest.mark.parametrize("reconn_delay", [0, 1])
def test_states_closed_happy_path(
    address: str,
    open_called: bool,
    rc: MQTTReasonCode | None,
    reconn_delay: int,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    decoder: Mock,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address(address), reconnect_delay=reconn_delay)
    fsm = FSM(env=env, init_state=ClosedState, error_state=ShutdownState)

    fsm.requested_state = ConnectingState
    state_data.disconnect_rc = rc
    state_data.open_called = open_called
    ClosedState.enter(fsm, state_data, env, params)
    if open_called:
        if rc is not None:
            expected_data = MQTTDisconnectPacket(rc).encode()
            mock_socket.send.assert_called_once()
            data = mock_socket.send.call_args[0][0]
            if params.address.is_websocket():
                assert data[0] == OpCode.BINARY.value | 0x80
                assert data[1] == len(expected_data) | 0x80
                mask = data[2:6]
                masked_data = data[6:]
                unmasked_data = apply_mask(mask, masked_data)
                assert unmasked_data == expected_data
            else:
                assert data == expected_data
    mock_socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
    mock_socket.close.assert_called_once()
    decoder.reset.assert_called_once()
    # We will assert that the write buffer is empty in another test.
    if reconn_delay > 0 and open_called:
        assert fsm.state is ReconnectWaitState
    else:
        assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("exc", [ssl.SSLWantReadError, ssl.SSLWantWriteError, BlockingIOError, OSError])
def test_states_closed_errors(
    exc: type[Exception],
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ClosedState, error_state=ShutdownState)

    fsm.previous_state = ClosingState
    state_data.open_called = True
    state_data.disconnect_rc = MQTTReasonCode.NormalDisconnection
    mock_socket.send.side_effect = exc("TEST")
    mock_socket.shutdown.side_effect = OSError("TEST")
    mock_socket.close.side_effect = OSError("TEST")
    ClosedState.enter(fsm, state_data, env, params)
    mock_socket.send.assert_called_once_with(MQTTDisconnectPacket().encode())
    mock_socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
    mock_socket.close.assert_called_once()
    assert state_data.open_called is False
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


def test_states_closed_write_buffer_not_empty(
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ClosedState, error_state=ShutdownState)

    fsm.previous_state = ConnectingState
    state_data.write_buffer.extend(b"test_data")
    ClosedState.enter(fsm, state_data, env, params)
    mock_socket.send.assert_not_called()
    assert len(state_data.write_buffer) == 0
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("close_exc", [OSError, None])
@pytest.mark.parametrize("open_called", [True, False])
def test_states_shutdown_enter(
    open_called: bool,
    close_exc: type[Exception] | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    decoder: Mock,
    mocker: MockerFixture,
    mock_socket: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"))
    fsm = FSM(env=env, init_state=ShutdownState, error_state=ShutdownState)

    mock_write_buffer = mocker.Mock()
    state_data.write_buffer = mock_write_buffer
    state_data.open_called = open_called
    mock_socket.close.side_effect = close_exc("TEST") if close_exc else None
    ShutdownState.enter(fsm, state_data, env, params)
    if open_called:
        pass
    mock_socket.close.assert_called_once()
    decoder.reset.assert_called_once()
    mock_write_buffer.clear.assert_called_once()
    assert fsm.state is ShutdownState

    callbacks.assert_not_called()


@pytest.mark.parametrize(
    "address", [
        "ws://test_host:1234",
        "wss://test_host:1234",
    ],
)
@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_websocket_handshake_request_happy_path(
    address: str,
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: Mock,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
    mocker: MockerFixture,
) -> None:
    mock_write_buffer = mocker.Mock()
    state_data.write_buffer = mock_write_buffer
    params = ConnectParams(address=Address(address))
    fsm = FSM(env=env, init_state=WebsocketHandshakeRequestState, error_state=ShutdownState)
    mock_timeout.get_timeout.return_value = 1 if max_wait is None else max_wait

    # Enter state.
    WebsocketHandshakeRequestState.enter(fsm, state_data, env, params)
    assert state_data.ws_nonce
    mock_write_buffer.clear.assert_called_once()
    mock_write_buffer.extend.assert_called_once()
    data = mock_write_buffer.extend.call_args[0][0]
    mock_write_buffer.reset_mock()
    state_data.write_buffer = bytearray(data)
    body = data.decode()
    assert "GET / HTTP/1.1\r" in body
    assert "Host: test_host:1234\r" in body
    assert "Upgrade: websocket" in body
    assert "Connection: Upgrade" in body
    assert f"Sec-WebSocket-Key: {state_data.ws_nonce}" in body
    assert "Sec-WebSocket-Version: 13" in body
    assert fsm.state is WebsocketHandshakeRequestState

    # First handle, blocked write.
    mock_socket.send.side_effect = BlockingIOError
    ret = WebsocketHandshakeRequestState.handle(fsm, state_data, env, params, max_wait=max_wait)
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert ret is False
    if max_wait:
        mock_select.select.assert_called_once_with(write=True, timeout=1)
        mock_select.reset_mock()

    # Second handle, partial write.
    mock_socket.send.side_effect = None
    mock_socket.send.return_value = 10
    ret = WebsocketHandshakeRequestState.handle(fsm, state_data, env, params, max_wait=max_wait)
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert ret is False
    if max_wait:
        mock_select.select.assert_called_once_with(write=True, timeout=1)
        mock_select.reset_mock()
    # Check remaining contents of write buffer.
    assert state_data.write_buffer == bytearray(data[10:])

    # Third handle, complete.
    mock_socket.send.side_effect = None
    mock_socket.send.return_value = len(data) - 10
    ret = WebsocketHandshakeRequestState.handle(fsm, state_data, env, params, max_wait=max_wait)
    mock_socket.send.assert_called_once_with(state_data.write_buffer)
    mock_socket.reset_mock()
    assert ret is True

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_websocket_handshake_request_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=WebsocketHandshakeRequestState, error_state=ShutdownState)

    fsm.previous_state = ConnectingState
    WebsocketHandshakeRequestState.enter(fsm, state_data, env, params)
    mock_timeout.exceeded.return_value = True
    ret = WebsocketHandshakeRequestState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_websocket_handshake_request_no_send(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("mqtt://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=WebsocketHandshakeRequestState, error_state=ShutdownState)

    fsm.previous_state = ConnectingState
    WebsocketHandshakeRequestState.enter(fsm, state_data, env, params)
    mock_socket.send.return_value = 0
    ret = WebsocketHandshakeRequestState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_websocket_handshake_response_happy_path(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_socket: Mock,
    mock_timeout: Mock,
) -> None:
    params = ConnectParams(address=Address("ws://testhost"))
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)
    state_data.ws_nonce = generate_nonce()
    expected_key = generate_handshake_key(state_data.ws_nonce)

    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)
    mock_socket.recv.return_value = (
        "HTTP/1.1 101 Switching Protocols\r\n"
        "Upgrade: websocket\r\n"
        "Connection: Upgrade\r\n"
        f"Sec-WebSocket-Accept: {expected_key}\r\n"
        "Sec-WebSocket-Protocol: mqtt\r\n"
        "\r\n"
    ).encode()
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is MQTTHandshakeConnectState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_websocket_handshake_response_invalid(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
) -> None:
    params = ConnectParams(address=Address("ws://testhost"))
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)
    state_data.ws_nonce = generate_nonce()
    expected_key = generate_handshake_key(state_data.ws_nonce)

    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)

    # Invalid response status line.
    mock_socket.recv.return_value = (
        "HTTP/1.1 400 Bad Request\r\n"
        "Upgrade: websocket\r\n"
        "Connection: Upgrade\r\n"
        f"Sec-WebSocket-Accept: {expected_key}\r\n"
        "Sec-WebSocket-Protocol: mqtt\r\n"
        "\r\n"
    ).encode()
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)
    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)
    mock_socket.reset_mock()

    # Missing Sec-WebSocket-Accept header.
    mock_socket.recv.return_value = (
        "HTTP/1.1 101 Switching Protocols\r\n"
        "Upgrade: websocket\r\n"
        "Connection: Upgrade\r\n"
        "\r\n"
    ).encode()
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)
    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)
    mock_socket.reset_mock()

    # Invalid Sec-WebSocket-Accept header.
    mock_socket.recv.return_value = (
        "HTTP/1.1 101 Switching Protocols\r\n"
        "Upgrade: websocket\r\n"
        "Connection: Upgrade\r\n"
        "Sec-WebSocket-Accept: invalid_key\r\n"
        "\r\n"
    ).encode()
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)
    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)
    mock_socket.reset_mock()

    # Missing Sec-WebSocket-Protocol header.
    mock_socket.recv.return_value = (
        "HTTP/1.1 101 Switching Protocols\r\n"
        "Upgrade: websocket\r\n"
        "Connection: Upgrade\r\n"
        f"Sec-WebSocket-Accept: {expected_key}\r\n"
        "\r\n"
    ).encode()
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)
    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)
    mock_socket.reset_mock()

    # Invalid Sec-WebSocket-Protocol header.
    mock_socket.recv.return_value = (
        "HTTP/1.1 101 Switching Protocols\r\n"
        "Upgrade: websocket\r\n"
        "Connection: Upgrade\r\n"
        f"Sec-WebSocket-Accept: {expected_key}\r\n"
        "Sec-WebSocket-Protocol: invalid_protocol\r\n"
        "\r\n"
    ).encode()
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)
    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)
    mock_socket.reset_mock()

    # Incomplete response.
    mock_socket.recv.return_value = "HTTP/1.1".encode()
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    assert fsm.state is WebsocketHandshakeResponseState
    if max_wait:
        mock_select.select.assert_called_once_with(read=True, timeout=1)

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
def test_states_websocket_handshake_response_timeout(
    max_wait: float | None,
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_timeout: Mock
) -> None:
    params = ConnectParams(address=Address("ws://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)

    fsm.previous_state = WebsocketHandshakeRequestState
    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)
    mock_timeout.exceeded.return_value = True
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is True
    assert fsm.state is ClosedState

    callbacks.assert_not_called()


@pytest.mark.parametrize("max_wait", [None, 0.0])
@pytest.mark.parametrize("exc", [ssl.SSLWantReadError, ssl.SSLWantWriteError, BlockingIOError])
def test_states_websocket_handshake_response_errors(
    max_wait: float | None,
    exc: type[Exception],
    callbacks: EnvironmentCallbacks,
    state_data: StateData,
    env: StateEnvironment,
    mock_select: Mock,
    mock_socket: Mock,
    mock_timeout: Mock,
) -> None:
    params = ConnectParams(address=Address("ws://testhost"), connect_timeout=5)
    fsm = FSM(env=env, init_state=WebsocketHandshakeResponseState, error_state=ShutdownState)

    fsm.previous_state = WebsocketHandshakeRequestState
    WebsocketHandshakeResponseState.enter(fsm, state_data, env, params)

    mock_select.select.return_value = (True, False)
    mock_socket.recv.side_effect = exc("TEST")
    ret = WebsocketHandshakeResponseState.handle(fsm, state_data, env, params, max_wait)
    assert ret is False
    assert fsm.state is WebsocketHandshakeResponseState

    callbacks.assert_not_called()



import socket
import ssl
from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection.decoder import ClosedSocketError
from ohmqtt.connection.ws_decoder import WebsocketDecoder
from ohmqtt.connection.wslib import WebsocketError, OpCode


@pytest.fixture
def mock_socket(mocker: MockerFixture) -> Mock:
    """Return a mock socket for testing."""
    return mocker.Mock(spec=socket.socket)  # type: ignore[no-any-return]


def test_ws_decoder_defaults() -> None:
    """reset() should set the decoder to its default state."""
    decoder = WebsocketDecoder()
    decoder2 = WebsocketDecoder()
    decoder2.reset()
    assert decoder.opcode == decoder2.opcode
    assert decoder.length == decoder2.length
    assert decoder.data == decoder2.data


@pytest.mark.parametrize("payload_length", [0, 125, 126, 127, 255, 65535, 65536])
def test_ws_decoder_payload_lengths(payload_length: int, mock_socket: Mock) -> None:
    """Test decoding frames with various payload lengths."""
    decoder = WebsocketDecoder()

    payload = b"x" * payload_length
    if payload_length <= 125:
        header = bytes([0x80 | OpCode.BINARY, payload_length])
    elif payload_length <= 0xFFFF:
        header = bytes([0x80 | OpCode.BINARY, 126]) + payload_length.to_bytes(2, "big")
    else:
        header = bytes([0x80 | OpCode.BINARY, 127]) + payload_length.to_bytes(8, "big")

    mock_socket.recv.side_effect = [bytes([b]) for b in header] + [payload]
    out = decoder.decode(mock_socket)
    assert out == (OpCode.BINARY, payload)


@pytest.mark.parametrize("exc", [BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError])
def test_ws_decoder_drip_feed(exc: Exception, mock_socket: Mock) -> None:
    """Feed the decoder one byte at a time."""
    decoder = WebsocketDecoder()
    # Use a packet with a payload of 255 bytes to ensure we have to read
    # multiple bytes for the frame length.
    payload = b"x" * 255
    frame = bytes([0x80 | OpCode.BINARY]) + bytes([126, 0, 255]) + payload

    # Setup mock to return one byte at a time
    for i in range(len(frame) - 1):
        mock_socket.recv.side_effect = [frame[i:i+1], exc]
        assert decoder.decode(mock_socket) is None

    # Final byte completes the packet
    mock_socket.recv.side_effect = [frame[-1:]]
    out = decoder.decode(mock_socket)
    assert out == (OpCode.BINARY, payload)


@pytest.mark.parametrize("exc", [BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError])
@pytest.mark.parametrize("closer", [b"", OSError])
@pytest.mark.parametrize("available_bytes", [n for n in range(6)])
def test_ws_decoder_drip_partial_closures(exc: Exception, closer: bytes | Exception,
                                          available_bytes: int, mock_socket: Mock) -> None:
    """Feed the decoder partial data, then close the socket."""
    decoder = WebsocketDecoder()
    # Use a packet with a payload of 255 bytes to ensure we have to read
    # multiple bytes for the frame length.
    payload = b"x" * 255
    frame = bytes([0x80 | OpCode.BINARY]) + bytes([126, 0, 255]) + payload

    # Send a few bytes then simulate closed socket
    if available_bytes > 0:
        mock_socket.recv.side_effect = [bytes([b]) for b in frame[:available_bytes]] + [exc]
    else:
        mock_socket.recv.side_effect = [exc]

    assert decoder.decode(mock_socket) is None

    mock_socket.recv.side_effect = [closer]
    with pytest.raises(ClosedSocketError):
        decoder.decode(mock_socket)


def test_ws_decoder_protocol_errors(mock_socket: Mock) -> None:
    """Test that protocol errors raise WebsocketError."""
    decoder = WebsocketDecoder()

    # Fragmented frame (FIN bit not set)
    mock_socket.recv.side_effect = [bytes([OpCode.BINARY])]
    with pytest.raises(WebsocketError, match="Fragmented WebSocket frames are not supported"):
        decoder.decode(mock_socket)

    decoder = WebsocketDecoder()

    # Reserved bits set
    mock_socket.recv.side_effect = [bytes([0x80 | 0x70 | OpCode.BINARY])]
    with pytest.raises(WebsocketError, match="Reserved WebSocket bits are set"):
        decoder.decode(mock_socket)

    decoder = WebsocketDecoder()

    # Invalid opcode
    mock_socket.recv.side_effect = [bytes([0x80 | 0x03])]
    with pytest.raises(WebsocketError, match="Invalid WebSocket opcode"):
        decoder.decode(mock_socket)

    decoder = WebsocketDecoder()

    # Masked frame from server
    mock_socket.recv.side_effect = [bytes([0x80 | OpCode.BINARY]), b"\x80"]
    with pytest.raises(WebsocketError, match="Masked frames from server are not allowed"):
        decoder.decode(mock_socket)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import ClassVar, Sequence, TYPE_CHECKING

from ..mqtt_spec import MQTTPacketType

if TYPE_CHECKING:
    from ..property import MQTTProperties


class MQTTPacket(metaclass=ABCMeta):
    """Base class for MQTT packets."""
    packet_type: ClassVar[MQTTPacketType]
    props_type: ClassVar[type[MQTTProperties]]
    __slots__: Sequence[str] = tuple()

    @abstractmethod
    def __str__(self) -> str:
        """Render the packet metadata as a human-readable string."""

    @abstractmethod
    def encode(self) -> bytes:
        """Encode the packet to bytes."""

    @classmethod
    @abstractmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPacket:
        """Decode a packet from bytes."""



from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode


def test_mqtt_error() -> None:
    e = MQTTError("test", MQTTReasonCode.ProtocolError)
    assert str(e) == "test (reason code: ProtocolError{0x82})"
    assert e.reason_code == MQTTReasonCode.ProtocolError
    assert isinstance(e, Exception)



# Copyright 2018 Simon Davy
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# WARNING: the code in the gist generates self-signed certs, for the purposes of testing in development.
# Do not use these certs in production, or You Will Have A Bad Time.
#
# Caveat emptor
#

from datetime import datetime, timedelta, timezone
import ipaddress
from typing import Any, Sequence


def generate_selfsigned_cert(
    hostname: str,
    ip_addresses: Sequence[str] | None = None,
    key: Any = None,
) -> tuple[bytes, bytes]:
    """Generates self signed certificate for a hostname, and optional IP addresses."""
    from cryptography import x509
    from cryptography.x509.oid import NameOID
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    # Generate our key
    if key is None:
        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend(),
        )

    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, hostname)
    ])

    # best practice seem to be to include the hostname in the SAN, which *SHOULD* mean COMMON_NAME is ignored.
    alt_names: list[x509.DNSName | x509.IPAddress] = [x509.DNSName(hostname)]

    # allow addressing by IP, for when you don't have real DNS (common in most testing scenarios
    if ip_addresses:
        for addr in ip_addresses:
            # openssl wants DNSnames for ips...
            alt_names.append(x509.DNSName(addr))
            # ... whereas golang's crypto/tls is stricter, and needs IPAddresses
            # note: older versions of cryptography do not understand ip_address objects
            alt_names.append(x509.IPAddress(ipaddress.ip_address(addr)))

    san = x509.SubjectAlternativeName(alt_names)

    # path_len=0 means this cert can only sign itself, not other certs.
    basic_constraints = x509.BasicConstraints(ca=True, path_length=0)
    now = datetime.now(timezone.utc)
    cert = (
        x509.CertificateBuilder()
        .subject_name(name)
        .issuer_name(name)
        .public_key(key.public_key())
        .serial_number(1000)
        .not_valid_before(now)
        .not_valid_after(now + timedelta(days=10*365))
        .add_extension(basic_constraints, False)
        .add_extension(san, False)
        .sign(key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
    key_pem = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return cert_pem, key_pem



import socket
from typing import TypedDict

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection.address import Address
from ohmqtt.platform import AF_UNIX, HAS_AF_UNIX, PlatformError


class AddressTestCase(TypedDict, total=False):
    address: str
    scheme: str
    family: str
    host: str
    port: int
    username: str | None
    password: str | None
    use_tls: bool


def lookup_family(family: str) -> socket.AddressFamily:
    """Convert family string to socket address family."""
    if family == "AF_INET":
        return socket.AF_INET
    if family == "AF_INET6":
        return socket.AF_INET6
    if family == "AF_UNIX" and HAS_AF_UNIX:
        return AF_UNIX
    raise ValueError(f"Unsupported address family: {family}")


def test_address_valid(test_data: list[AddressTestCase]) -> None:
    """Test the Address class with a Unix socket address."""
    for case in test_data:
        case_addr = case["address"]
        try:
            address = Address(case_addr)
        except Exception as exc:
            raise Exception(f"Failed to parse address {case_addr}") from exc
        assert address.scheme == case["scheme"], f"scheme for {case_addr}"
        assert address.family == lookup_family(case["family"]), f"family for {case_addr}"
        assert address.host == case["host"], f"host for {case_addr}"
        assert address.port == case["port"], f"port for {case_addr}"
        assert address.username == case.get("username", None), f"username for {case_addr}"
        assert address.password == case.get("password", None), f"password for {case_addr}"
        assert address.use_tls is case.get("use_tls", False), f"use_tls for {case_addr}"
        assert address.path == case.get("path", ""), f"path for {case_addr}"
        assert repr(address)
        if case.get("password", None) is not None:
            assert str(case["password"]) not in repr(address), f"password not hidden for {case_addr}"


@pytest.mark.skipif(
    not HAS_AF_UNIX,
    reason="Unix domain sockets are not available on this platform",
)
def test_address_unix(test_data: list[AddressTestCase]) -> None:
    test_address_valid(test_data)


def test_address_unix_unsupported(mocker: MockerFixture, test_data_file: dict[str, list[AddressTestCase]]) -> None:
    """Test a platform which does not support Unix sockets."""
    mocker.patch("ohmqtt.connection.address.HAS_AF_UNIX", new=False)
    for case in test_data_file["test_address_unix"]:
        case_addr = case["address"]
        with pytest.raises(PlatformError):
            Address(case_addr)


def test_address_invalid(test_data: list[AddressTestCase]) -> None:
    """Test the Address class with invalid addresses."""
    for case in test_data:
        try:
            Address(case["address"])
        except (PlatformError, ValueError):
            pass
        else:
            pytest.fail(f"Expected ValueError for address: {case['address']}")


def test_address_empty() -> None:
    """Test the Address class with an empty address.

    This should result in an Address object with no values set."""
    address = Address("")
    for attr in Address.__slots__:
        assert not hasattr(address, attr), attr
    assert not address.use_tls
    repr(address)


def test_address_is_websocket() -> None:
    """Test the is_websocket property of the Address class."""
    ws_address = Address("ws://example.com/path")
    assert ws_address.is_websocket()
    wss_address = Address("wss://example.com/path")
    assert wss_address.is_websocket()
    non_ws_address = Address("mqtt://example.com:1883")
    assert not non_ws_address.is_websocket()



import pytest

from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTPropertyId, MQTTReasonCode
from ohmqtt.property import MQTTPublishProps
from ohmqtt.serialization import encode_varint, encode_uint16


def test_property_duplicate_key() -> None:
    data = bytearray(encode_varint(MQTTPropertyId.TopicAlias) + encode_uint16(1))
    data.extend(encode_varint(MQTTPropertyId.TopicAlias) + encode_uint16(2))
    data[0:0] = encode_varint(len(data))
    view = memoryview(data)
    view.toreadonly()

    with pytest.raises(MQTTError, match=r"Duplicate") as excinfo:
        MQTTPublishProps.decode(view)
    assert excinfo.value.reason_code == MQTTReasonCode.ProtocolError



import socket
import threading

import pytest

from ohmqtt.connection.selector import InterruptibleSelector


def test_selector_protection() -> None:
    selector = InterruptibleSelector()

    with pytest.raises(RuntimeError):
        selector.select()
    with pytest.raises(RuntimeError):
        selector.interrupt()


def test_selector_interrupt(loopback_socket: socket.socket) -> None:
    selector = InterruptibleSelector()

    start = threading.Event()
    def wait_for_interrupt() -> None:
        with selector:
            selector.change_sock(loopback_socket)
            start.set()
            selector.select(read=True)

    thread = threading.Thread(target=wait_for_interrupt, daemon=True)
    thread.start()

    start.wait(1.0)
    with selector:
        selector.interrupt()

    thread.join(1.0)
    assert not thread.is_alive()


def test_selector_change_sock() -> None:
    selector = InterruptibleSelector()
    sock1 = socket.socket()
    sock2 = socket.socket()

    with selector:
        selector.change_sock(sock1)
        selector.change_sock(sock2)
        selector.change_sock(sock2)  # Redundant change, should not raise


def test_selector_no_sock() -> None:
    selector = InterruptibleSelector()

    with selector:
        with pytest.raises(RuntimeError):
            selector.select()


def test_selector_close(loopback_socket: socket.socket) -> None:
    selector = InterruptibleSelector()

    with selector:
        selector.change_sock(loopback_socket)
        selector.close()
        with pytest.raises(RuntimeError):
            selector.change_sock(loopback_socket)
        with pytest.raises(RuntimeError):
            selector.select()
        with pytest.raises(RuntimeError):
            selector.interrupt()


def test_selector_select(loopback_socket: socket.socket) -> None:
    selector = InterruptibleSelector()

    with selector:
        selector.change_sock(loopback_socket)
        assert selector.select(read=True, timeout=0.1) == (False, False)
        assert selector.select(write=True, timeout=0.1) == (False, True)
        loopback_socket.test_sendall(b"test")  # type: ignore[attr-defined]
        assert selector.select(read=True, timeout=0.1) == (True, False)
        with pytest.raises(ValueError):
            selector.select(timeout=0.1)



"""E2E Tests"""

from __future__ import annotations

from collections import deque
import time
from typing import Final, Iterator

import pytest

from .util.fake_broker import FakeBroker, FakeWebsocketBroker
from ohmqtt.client import Client
from ohmqtt.connection import InvalidStateError
from ohmqtt.logger import get_logger
from ohmqtt.mqtt_spec import MQTTQoS
from ohmqtt.packet import (
    MQTTPacket,
    MQTTConnectPacket,
    MQTTSubscribePacket,
    MQTTUnsubscribePacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRelPacket,
    MQTTDisconnectPacket,
)

logger: Final = get_logger("tests.test_z2z")


@pytest.fixture
def broker() -> Iterator[FakeBroker]:
    broker = FakeBroker()
    with broker:
        yield broker


@pytest.fixture
def ws_broker() -> Iterator[FakeWebsocketBroker]:
    broker = FakeWebsocketBroker()
    with broker:
        yield broker


def get_client(broker: FakeBroker, scheme: str, db_path: str) -> Client:
    client = Client(db_path)
    client.start_loop()

    client.connect(
        address=f"{scheme}://localhost:{broker.port}",
        client_id="test_client",
        clean_start=True,
    )
    client.wait_for_connect(timeout=0.25)
    assert client.is_connected()
    assert broker.received.popleft() == MQTTConnectPacket(
        client_id="test_client",
        clean_start=True,
    )
    return client


def wait_for_queue_item(queue: deque[MQTTPacket], timeout: float = 0.1) -> object:
    t0 = time.monotonic()
    while True:
        if queue:
            return queue.popleft()
        if time.monotonic() > t0 + timeout:
            raise TimeoutError("Timed out waiting for queue item")
        time.sleep(0.001)


@pytest.mark.parametrize("scheme", ["mqtt", "ws"])
@pytest.mark.parametrize("db_path", ["", ":memory:"])
def test_z2z_happy_path(db_path: str, scheme: str) -> None:
    broker = FakeWebsocketBroker() if scheme == "ws" else FakeBroker()
    with broker:
        client = get_client(broker, scheme, db_path)
        delay: Final = 0.05  # seconds grace period

        client_received: deque[MQTTPacket] = deque()
        def callback(client: Client, packet: MQTTPacket) -> None:
            client_received.append(packet)

        # SUBSCRIBE
        sub_handle = client.subscribe("test/topic", callback)
        assert sub_handle is not None
        sub_handle.wait_for_ack(timeout=0.5)
        assert wait_for_queue_item(broker.received) == MQTTSubscribePacket(
            topics=[("test/topic", 2)],
            packet_id=1,
        )

        # PUBLISH QoS 0
        for _ in range(10):
            pub_handle = client.publish("test/topic", b"banana", qos=0)
            time.sleep(delay)
            assert wait_for_queue_item(broker.received) == wait_for_queue_item(client_received) == MQTTPublishPacket(
                topic="test/topic",
                payload=b"banana",
            )

        # PUBLISH QoS 1
        for _ in range(1, 10):
            pub_handle = client.publish("test/topic", b"coconut", qos=1)
            pub_handle.wait_for_ack(timeout=0.25)
            time.sleep(delay)
            broker_rec = wait_for_queue_item(broker.received)
            client_rec = wait_for_queue_item(client_received)
            assert isinstance(broker_rec, MQTTPublishPacket)
            assert isinstance(client_rec, MQTTPublishPacket)
            assert broker_rec.topic == client_rec.topic == "test/topic"
            assert broker_rec.payload == client_rec.payload == b"coconut"
            assert broker_rec.qos == client_rec.qos == MQTTQoS.Q1
            assert wait_for_queue_item(broker.received) == MQTTPubAckPacket(packet_id=broker_rec.packet_id)

        # UNSUBSCRIBE
        unsub_handle = client.unsubscribe("test/topic")
        assert unsub_handle is not None
        unsub_handle.wait_for_ack(timeout=0.5)
        time.sleep(delay)
        assert wait_for_queue_item(broker.received) == MQTTUnsubscribePacket(
            topics=["test/topic"],
            packet_id=1,
        )

        # PUBLISH QoS 2
        for _ in range(10, 20):
            pub_handle = client.publish("test/topic", b"pineapple", qos=2)
            pub_handle.wait_for_ack(timeout=0.25)
            time.sleep(delay)
            broker_rec = wait_for_queue_item(broker.received)
            assert isinstance(broker_rec, MQTTPublishPacket)
            assert broker_rec.topic == "test/topic"
            assert broker_rec.payload == b"pineapple"
            assert broker_rec.qos == MQTTQoS.Q2
            assert wait_for_queue_item(broker.received) == MQTTPubRelPacket(packet_id=broker_rec.packet_id)

        # DISCONNECT
        client.disconnect()
        client.wait_for_disconnect(timeout=0.25)
        time.sleep(delay)
        assert wait_for_queue_item(broker.received) == MQTTDisconnectPacket()

        client.shutdown()
        client.wait_for_shutdown(timeout=0.25)


@pytest.mark.parametrize("db_path", ["", ":memory:"])
def test_z2z_saturation_qos0(db_path: str, broker: FakeBroker) -> None:
    num: Final = 5000
    timeout: Final = 5.0  # seconds
    sz: Final = 1000
    delay: Final = 0.01  # seconds grace period
    client = get_client(broker, "mqtt", db_path)

    client_received: deque[MQTTPacket] = deque()
    def callback(client: Client, packet: MQTTPacket) -> None:
        client_received.append(packet)

    sub_handle = client.subscribe("test/topic", callback)
    assert sub_handle is not None
    sub_handle.wait_for_ack(timeout=0.5)
    assert broker.received.popleft() == MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    )

    for _ in range(num):
        client.publish("test/topic", b"x" * sz)

    t0 = time.monotonic()
    while True:
        if len(broker.received) == num and len(client_received) == num:
            break
        if time.monotonic() > t0 + timeout:
            raise TimeoutError(f"Didn't publish and receive messages in time {len(broker.received)}:{len(client_received)}")
        time.sleep(delay)

    for _ in range(num):
        broker_rec = broker.received.popleft()
        client_rec = client_received.popleft()
        assert isinstance(broker_rec, MQTTPublishPacket)
        assert isinstance(client_rec, MQTTPublishPacket)
        assert broker_rec == client_rec == MQTTPublishPacket(
            topic="test/topic",
            payload=b"x" * sz,
        )


@pytest.mark.parametrize("db_path", ["", ":memory:"])
def test_z2z_saturation_qos1(db_path: str, broker: FakeBroker) -> None:
    num: Final = 2000
    timeout: Final = 5.0  # seconds
    sz: Final = 1000
    delay: Final = 0.01  # seconds grace period
    client = get_client(broker, "mqtt", db_path)

    client_received: deque[MQTTPacket] = deque()
    def callback(client: Client, packet: MQTTPacket) -> None:
        client_received.append(packet)

    sub_handle = client.subscribe("test/topic", callback)
    assert sub_handle is not None
    sub_handle.wait_for_ack(timeout=0.5)
    assert broker.received.popleft() == MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    )

    for _ in range(num):
        client.publish("test/topic", b"x" * sz, qos=1)

    t0 = time.monotonic()
    while True:
        if len(broker.received) == num * 2 and len(client_received) == num:
            break
        if time.monotonic() > t0 + timeout:
            raise TimeoutError(f"Didn't publish and receive messages in time {len(broker.received)}:{len(client_received)}")
        time.sleep(delay)

    for _ in range(num):
        broker_rec = broker.received.popleft()
        while isinstance(broker_rec, MQTTPubAckPacket):
            broker_rec = broker.received.popleft()
        client_rec = client_received.popleft()
        assert isinstance(broker_rec, MQTTPublishPacket)
        assert isinstance(client_rec, MQTTPublishPacket)
        assert broker_rec.topic == client_rec.topic == "test/topic"
        assert broker_rec.payload == client_rec.payload == b"x" * sz
        assert broker_rec.qos == client_rec.qos == MQTTQoS.Q1


@pytest.mark.parametrize("db_path", ["", ":memory:"])
def test_z2z_saturation_qos2(db_path: str, broker: FakeBroker) -> None:
    num: Final = 1000
    timeout: Final = 5.0  # seconds
    sz: Final = 1000
    delay: Final = 0.01  # seconds grace period
    client = get_client(broker, "mqtt", db_path)

    client_received: deque[MQTTPacket] = deque()
    def callback(client: Client, packet: MQTTPacket) -> None:
        client_received.append(packet)

    sub_handle = client.subscribe("test/topic", callback)
    assert sub_handle is not None
    sub_handle.wait_for_ack(timeout=0.5)
    assert broker.received.popleft() == MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    )

    for _ in range(num):
        client.publish("test/topic", b"x" * sz, qos=2)

    t0 = time.monotonic()
    while True:
        if len(broker.received) == num * 4 and len(client_received) == num:
            break
        if time.monotonic() > t0 + timeout:
            raise TimeoutError(f"Didn't publish and receive messages in time {len(broker.received)}:{len(client_received)}")
        time.sleep(delay)

    for _ in range(num):
        broker_rec = broker.received.popleft()
        while not isinstance(broker_rec, MQTTPublishPacket):
            broker_rec = broker.received.popleft()
        client_rec = client_received.popleft()
        assert isinstance(broker_rec, MQTTPublishPacket)
        assert isinstance(client_rec, MQTTPublishPacket)
        assert broker_rec.topic == client_rec.topic == "test/topic"
        assert broker_rec.payload == client_rec.payload == b"x" * sz
        assert broker_rec.qos == client_rec.qos == MQTTQoS.Q2


def test_z2z_change_endpoint(broker: FakeBroker, ws_broker: FakeWebsocketBroker) -> None:
    with Client() as client:

        # Connect to TCP broker
        client.connect(
            address=f"mqtt://localhost:{broker.port}",
            client_id="test_client",
            clean_start=True,
        )
        client.wait_for_connect(timeout=0.25)
        assert client.is_connected()
        assert broker.received.popleft() == MQTTConnectPacket(
            client_id="test_client",
            clean_start=True,
        )

        # Try to connect to WebSocket broker
        with pytest.raises(InvalidStateError):
            client.connect(
                address=f"ws://localhost:{ws_broker.port}",
                client_id="test_client",
                clean_start=True,
            )

        # Disconnect and try again
        client.disconnect()
        client.wait_for_disconnect(timeout=0.25)
        assert client.is_connected() is False

        client.connect(
            address=f"ws://localhost:{ws_broker.port}",
            client_id="test_client",
            clean_start=True,
        )
        client.wait_for_connect(timeout=0.25)
        assert client.is_connected()
        assert ws_broker.received.popleft() == MQTTConnectPacket(
            client_id="test_client",
            clean_start=True,
        )



from dataclasses import dataclass
import socket
from typing import Final, Mapping
from urllib.parse import urlparse, ParseResult

from ..platform import AF_UNIX, HAS_AF_UNIX, PlatformError


DEFAULT_PORTS: Final[Mapping[str, int]] = {
    "mqtt": 1883,
    "mqtts": 8883,
    "unix": 0,
    "ws": 80,
    "wss": 443,
}


def is_ipv6(hostname: str) -> bool:
    """Check if the hostname is an IPv6 address."""
    try:
        socket.inet_pton(socket.AF_INET6, hostname)
        return True
    except socket.error:
        return False


def _get_family(parsed: ParseResult) -> socket.AddressFamily:
    """Get the address family based on the parsed URL scheme."""
    if HAS_AF_UNIX and parsed.scheme == "unix":
        return AF_UNIX
    if parsed.scheme in ("mqtt", "mqtts", "ws", "wss"):
        if not parsed.hostname:
            raise ValueError("Hostname is required for mqtt and mqtts schemes")
        if is_ipv6(parsed.hostname):
            return socket.AF_INET6
        return socket.AF_INET
    raise ValueError(f"Unsupported scheme: {parsed.scheme}")


@dataclass(slots=True, init=False, frozen=True, repr=False)
class Address:
    scheme: str
    family: socket.AddressFamily
    host: str
    port: int
    path: str
    username: str | None
    password: str | None

    def __init__(self, address: str = "") -> None:
        """Parse the address string into family, host, port, username, and password."""
        # Special case: empty address is allowed, but slots will be empty.
        if not address:
            return
        if address.startswith("unix:"):
            if not HAS_AF_UNIX:
                raise PlatformError("Unix socket support is not available on this platform")
        elif "//" not in address:
            # urlparse may choke on some network address we wish to support, unless we guarantee a //.
            address = "//" + address
        parsed = urlparse(address, scheme="mqtt")
        object.__setattr__(self, "scheme", parsed.scheme)
        object.__setattr__(self, "family", _get_family(parsed))
        object.__setattr__(self, "host", parsed.hostname or parsed.path)
        if not self.host:
            raise ValueError("No path in address")
        if HAS_AF_UNIX and self.family == AF_UNIX and self.host == "/":
            raise ValueError("'/' is not a valid Unix socket path")
        object.__setattr__(self, "port", parsed.port if parsed.port is not None else DEFAULT_PORTS[parsed.scheme])
        default_path = "/" if self.scheme in ("ws", "wss") else ""
        object.__setattr__(self, "path", parsed.path if parsed.path and parsed.hostname else default_path)
        object.__setattr__(self, "username", parsed.username)
        object.__setattr__(self, "password", parsed.password)

    def __repr__(self) -> str:
        """Return a string representation of the address."""
        if not hasattr(self, "scheme"):
            # Handle the empty case.
            return "Address()"
        userpw = ""
        if self.username:
            userpw = self.username
        if self.password:
            userpw += ":<hidden>"
        if userpw:
            host = f"{userpw}@{self.host}"
        else:
            host = self.host
        if self.scheme != "unix":
            host = f"{host}:{self.port}"
        if self.path:
            host = f"{host}{self.path}"
        return f"Address({self.scheme}://{host})"

    @property
    def use_tls(self) -> bool:
        """Check if the address uses TLS."""
        return getattr(self, "scheme", None) in ("mqtts", "wss")

    def is_default_port(self) -> bool:
        """Check if the port is the default for the scheme."""
        return getattr(self, "port", None) == DEFAULT_PORTS.get(getattr(self, "scheme", ""), None)

    def is_websocket(self) -> bool:
        """Check if the address uses WebSocket."""
        return getattr(self, "scheme", None) in ("ws", "wss")



import contextlib
import nox
import sqlite3
import sys

nox.options.default_venv_backend = "uv|virtualenv"

PYPROJECT = nox.project.load_toml("pyproject.toml")

ALL_PYTHONS = [
    cls.split()[-1]
    for cls in PYPROJECT["project"]["classifiers"]
    if cls.startswith("Programming Language :: Python :: 3.")
]

# Include free threading builds where available
FT_PYTHONS = ["3.14"]
for py in FT_PYTHONS:
    if py in ALL_PYTHONS:
        ALL_PYTHONS.append(f"{py}t")


@nox.session(python=ALL_PYTHONS)
def tests(session: nox.Session) -> None:
    free_threading = session.python.endswith("t")
    coverage_file = f".coverage.{sys.platform}.{session.python}"
    pytest_env = {
        "COVERAGE_FILE": coverage_file,
    }

    session.install(".")
    session.install("--group", "dev")
    session.run("ruff", "check")
    session.run("typos")
    session.run("mypy")
    if free_threading:
        session.log("Skipping complexipy on free threading Python.")
    else:
        session.run("complexipy")
    session.run("pytest", env=pytest_env)

    if sys.platform.startswith("win"):
        with contextlib.closing(sqlite3.connect(coverage_file)) as con, con:
            con.execute("UPDATE file SET path = REPLACE(path, '\\', '/')")
            con.execute("DELETE FROM file WHERE SUBSTR(path, 2, 1) == ':'")


@nox.session(default=False)
def cover(session: nox.Session) -> None:
    """Coverage analysis."""
    session.install("coverage[toml]>=7.3")
    session.run("coverage", "combine")
    session.run("coverage", "report", "--fail-under=100", "--show-missing")
    session.run("coverage", "erase")



"""Constants from the MQTT specification."""

from enum import IntEnum
from typing import Final


MAX_PACKET_ID: Final = 0xffff


class MQTTPacketType(IntEnum):
    """Types of MQTT control packets, mapped to identifiers from the specification."""
    CONNECT = 0x01
    CONNACK = 0x02
    PUBLISH = 0x03
    PUBACK = 0x04
    PUBREC = 0x05
    PUBREL = 0x06
    PUBCOMP = 0x07
    SUBSCRIBE = 0x08
    SUBACK = 0x09
    UNSUBSCRIBE = 0x0a
    UNSUBACK = 0x0b
    PINGREQ = 0x0c
    PINGRESP = 0x0d
    DISCONNECT = 0x0e
    AUTH = 0x0f


class MQTTPropertyId(IntEnum):
    """Types of MQTT properties, mapped to identifiers from the specification."""
    PayloadFormatIndicator = 0x01
    MessageExpiryInterval = 0x02
    ContentType = 0x03
    ResponseTopic = 0x08
    CorrelationData = 0x09
    SubscriptionIdentifier = 0x0b
    SessionExpiryInterval = 0x11
    AssignedClientIdentifier = 0x12
    ServerKeepAlive = 0x13
    AuthenticationMethod = 0x15
    AuthenticationData = 0x16
    RequestProblemInformation = 0x17
    WillDelayInterval = 0x18
    RequestResponseInformation = 0x19
    ResponseInformation = 0x1a
    ServerReference = 0x1c
    ReasonString = 0x1f
    ReceiveMaximum = 0x21
    TopicAliasMaximum = 0x22
    TopicAlias = 0x23
    MaximumQoS = 0x24
    RetainAvailable = 0x25
    UserProperty = 0x26
    MaximumPacketSize = 0x27
    WildcardSubscriptionAvailable = 0x28
    SubscriptionIdentifierAvailable = 0x29
    SharedSubscriptionAvailable = 0x2a


class MQTTReasonCode(IntEnum):
    """Indicates the result of an operation."""
    Success = 0x00
    NormalDisconnection = 0x00
    GrantedQoS0 = 0x00
    GrantedQoS1 = 0x01
    GrantedQoS2 = 0x02
    DisconnectWithWillMessage = 0x04
    NoMatchingSubscribers = 0x10
    NoSubscriptionExisted = 0x11
    ContinueAuthentication = 0x18
    ReAuthenticate = 0x19
    UnspecifiedError = 0x80
    MalformedPacket = 0x81
    ProtocolError = 0x82
    ImplementationSpecificError = 0x83
    UnsupportedProtocolVersion = 0x84
    ClientIdentifierNotValid = 0x85
    BadUserNameOrPassword = 0x86
    NotAuthorized = 0x87
    ServerUnavailable = 0x88
    ServerBusy = 0x89
    Banned = 0x8A
    ServerShuttingDown = 0x8B
    BadAuthenticationMethod = 0x8C
    KeepAliveTimeout = 0x8D
    SessionTakenOver = 0x8E
    TopicFilterInvalid = 0x8F
    TopicNameInvalid = 0x90
    PacketIdentifierInUse = 0x91
    PacketIdentifierNotFound = 0x92
    ReceiveMaximumExceeded = 0x93
    TopicAliasInvalid = 0x94
    PacketTooLarge = 0x95
    MessageRateTooHigh = 0x96
    QuotaExceeded = 0x97
    AdministrativeAction = 0x98
    PayloadFormatInvalid = 0x99
    RetainNotSupported = 0x9A
    QoSNotSupported = 0x9B
    UseAnotherServer = 0x9C
    ServerMoved = 0x9D
    SharedSubscriptionsNotSupported = 0x9E
    ConnectionRateExceeded = 0x9F
    MaximumConnectTime = 0xA0
    SubscriptionIdentifiersNotSupported = 0xA1
    WildcardSubscriptionsNotSupported = 0xA2

    def is_error(self) -> bool:
        """Check if the reason code indicates an error."""
        return self >= 0x80


class MQTTQoS(IntEnum):
    """Quality of Service level for messaging."""
    Q0 = 0
    Q1 = 1
    Q2 = 2



"""Command line interface for ohmqtt."""

import argparse
import logging

from .publish import PublishCommand
from .subscribe import SubscribeCommand
from .. import __version__


def main(args: list[str]) -> None:
    """Main entry point for the ohmqtt CLI."""
    parser = argparse.ArgumentParser(description="ohmqtt Command Line Interface")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="enable verbose output",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="show the version of ohmqtt and exit",
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    PublishCommand.register(subparsers)
    SubscribeCommand.register(subparsers)

    parsed = parser.parse_args(args)

    if parsed.verbose:
        logging.basicConfig(level=logging.DEBUG)

    if hasattr(parsed, "func"):
        parsed.func(parsed)
    else:
        parser.print_help()



#!/usr/bin/env python3
"""Examples of authentication.  Three relevant features:

* TLS connection.
    Specify that you want to use TLS by using protocol mqtts:// in the server address.
    This authenticates the server and encrypts all MQTT traffic.

* Simple username/password in CONNECT.
    Specify the username and password in the server address.
    Use of TLS is strongly encouraged, otherwise the authenticating information can be captured and replayed.

* AUTH control packet exchange.
    Exchange AUTH packets to implement challenge/response authentication.
"""

import threading
from typing import Any

from ohmqtt import Client, MQTTAuthPacket, MQTTReasonCode


class AuthClient(Client):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.auth_done = threading.Event()

    def handle_auth(self, packet: MQTTAuthPacket) -> None:
        """Handle an AUTH packet from the server.

        You may send back AUTH packets from this handler."""
        method = packet.properties.AuthenticationMethod
        data = packet.properties.AuthenticationData
        rc = packet.reason_code
        print(f"auth {rc=} {method=} {data=}")
        if rc == MQTTReasonCode.ContinueAuthentication:
            self.auth(authentication_method=method, authentication_data=b"some auth data")
        else:
            self.auth_done.set()


def main() -> None:
    with AuthClient() as client:
        client.connect(
            # Use mqtts:// to specify TLS.
            # Use user:password@ to specify username and password.
            # Password will be encoded with UTF-8.
            "mqtts://user:password@localhost:8883",
            # Optionally provide your own TLS context.
            tls_context=None,
            # Optionally specify a hostname other than the one in the address for TLS.
            tls_hostname="localhost",
        )
        client.auth_done.wait(timeout=10)


if __name__ == "__main__":
    main()



import re


def validate_topic(topic: str) -> None:
    """Validate an MQTT topic name.

    Raises ValueError if the topic name is invalid."""
    if len(topic) == 0:
        raise ValueError("Topic name cannot be empty")
    if "\u0000" in topic:
        raise ValueError("Topic name cannot contain null characters")
    if "#" in topic or "+" in topic:
        raise ValueError("Topic name cannot contain wildcards '#' or '+'")
    if len(topic.encode("utf-8")) > 65535:
        raise ValueError("Topic name is too long (> 65535 bytes encoded as UTF-8)")


def validate_topic_filter(topic_filter: str) -> None:
    """Validate an MQTT topic filter.

    Raises ValueError if the topic filter is invalid."""
    if len(topic_filter) == 0:
        raise ValueError("Topic filter cannot be empty")
    if "\u0000" in topic_filter:
        raise ValueError("Topic filter cannot contain null characters")
    if len(topic_filter.encode("utf-8")) > 65535:
        raise ValueError("Topic filter is too long (> 65535 bytes encoded as UTF-8)")
    multi_level_wildcard_index = topic_filter.find("#")
    if multi_level_wildcard_index != -1:
        if multi_level_wildcard_index != len(topic_filter) - 1:
            raise ValueError("Multi-level wildcard '#' must be the last character in the topic filter")
        if len(topic_filter) > 1 and topic_filter[-2] != "/":
            raise ValueError("Multi-level wildcard '#' must be preceded by a '/' unless it is the only character in the topic filter")


def validate_share_name(share_name: str) -> None:
    """Validate an MQTT shared subscription name.

    Raises ValueError if the share name is invalid."""
    if len(share_name) == 0:
        raise ValueError("Share name cannot be empty")
    if "\u0000" in share_name:
        raise ValueError("Share name cannot contain null characters")
    if re.search(r"[+#/]", share_name):
        raise ValueError("Share name cannot contain characters '#' or '+' or '/'")
    if len(share_name.encode("utf-8")) > 65535:
        raise ValueError("Share name is too long (> 65535 bytes encoded as UTF-8)")


def _check_multi_level_wildcard(topic_filter: str, topic: str) -> bool:
    """Check if the topic matches the filter with a multi-level wildcard."""
    if len(topic_filter) == 1 and not topic.startswith("$"):
        return True  # Matches everything that doesn't start with '$'.
    # Otherwise the # must come at the end of the filter.
    # Match the path up to the last '/' in the filter.
    base = topic_filter[:-2]
    if base and topic.startswith(base):
        return True
    return False


def _check_single_level_wildcard(topic_filter: str, topic: str) -> bool:
    if topic_filter.startswith("+") and topic.startswith("$"):
        return False
    filter_levels = topic_filter.split("/")
    topic_levels = topic.split("/")
    if len(filter_levels) != len(topic_levels):
        return False
    for filter_level, topic_level in zip(filter_levels, topic_levels):
        if filter_level != "+" and filter_level != topic_level:
            return False
    return True


def match_topic_filter(topic_filter: str, topic: str) -> bool:
    """Check if the topic matches the filter.

    This method will validate the topic, but assumes that the filter is already validated."""
    validate_topic(topic)
    if topic_filter == topic:
        return True
    if "#" in topic_filter and _check_multi_level_wildcard(topic_filter, topic):
        return True
    if "+" in topic_filter and _check_single_level_wildcard(topic_filter, topic):
        return True
    return False


def join_share(topic_filter: str, share_name: str | None) -> str:
    """Join a topic filter with a share name."""
    return f"$share/{share_name}/{topic_filter}" if share_name is not None else topic_filter
